/*
Metric
Copyright (C) 2006-2011 Yangli Hector Yee
Copyright (C) 2011-2016 Steven Myint, Jeff Terrace

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place, Suite 330, Boston, MA 02111-1307 USA
*/

#ifndef PERCEPTUALDIFF_EXCEPTIONS_H
#define PERCEPTUALDIFF_EXCEPTIONS_H

#include <stdexcept>


namespace pdiff
{
    class PerceptualDiffException : public virtual std::invalid_argument
    {
    public:

        explicit PerceptualDiffException(const std::string &message)
            : std::invalid_argument(message)
        {
        }
    };
}

#endif
